use std::cmp::Ordering;
use std::collections::BTreeSet;
use std::io;
use std::iter::{repeat, successors};

#[derive(Clone, Copy, Debug, Eq)]
struct Coordinate {
    step: usize,
    x: i32,
    y: i32,
}

impl PartialEq for Coordinate {
    fn eq(&self, other: &Self) -> bool {
        self.x == other.x && self.y == other.y
    }
}

impl Ord for Coordinate {
    fn cmp(&self, other: &Self) -> Ordering {
        self.x.cmp(&other.x)
    }
}

impl PartialOrd for Coordinate {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        self.y.partial_cmp(&other.y)
    }
}

impl Coordinate {
    fn new() -> Self {
        Self {
            step: 0,
            x: 0,
            y: 0,
        }
    }

    fn next(self: &Self, direction: &str) -> Coordinate {
        match direction {
            "R" => Coordinate {
                step: self.step + 1,
                x: self.x + 1,
                y: self.y,
            },
            "L" => Coordinate {
                step: self.step + 1,
                x: self.x - 1,
                y: self.y,
            },
            "U" => Coordinate {
                step: self.step + 1,
                x: self.x,
                y: self.y + 1,
            },
            "D" => Coordinate {
                step: self.step + 1,
                x: self.x,
                y: self.y - 1,
            },
            _ => panic!("Wrong direction format!"),
        }
    }
}

fn direction_length(entry: String) -> (String, usize) {
    (
        entry.chars().nth(0).unwrap().to_string(),
        entry[1..].parse().unwrap(),
    )
}

fn to_coordinates(entries: Vec<String>) -> BTreeSet<Coordinate> {
    let mut wire = BTreeSet::new();

    let mut it = entries
        .into_iter()
        .map(direction_length)
        .flat_map(|(direction, length)| repeat(direction).take(length));

    successors(
        it.next()
            .map(|direction| (Coordinate::next(&Coordinate::new(), direction.as_ref()))),
        |c| it.next().map(|direction| (c.next(direction.as_ref()))),
    )
    .for_each(|c| {
        wire.insert(c);
    });

    wire
}

fn closest(input: String) -> Option<i32> {
    let entries: Vec<BTreeSet<Coordinate>> = input
        .trim()
        .split("\n")
        .map(|line| line.split(",").map(String::from).collect())
        .map(to_coordinates)
        .collect();

    println!(
        "{:?}",
        entries[0]
            .intersection(&entries[1])
            .cloned()
            .map(|c| c.x.abs() + c.y.abs())
            .collect::<Vec<i32>>()
    );

    entries[0]
        .intersection(&entries[1])
        .cloned()
        .map(|c| c.x.abs() + c.y.abs())
        .min()
}

fn best(input: String) -> Option<i32> {
    None
}

pub fn main() -> io::Result<()> {
    let closest = closest(std::include_str!("../day3-input.txt").to_owned());
    println!("{:?}", closest);

    Ok(())
}

#[test]
fn test_coordinates_for_entry() {
    let mut expected: BTreeSet<Coordinate> = BTreeSet::new();
    for i in 1..11 {
        expected.insert(Coordinate {
            step: i,
            x: i as i32,
            y: 0,
        });
    }
    assert_eq!(to_coordinates(vec!["R10".to_owned()]), expected);

    expected = BTreeSet::new();
    expected.insert(Coordinate {
        step: 1,
        x: 1,
        y: 0,
    });
    expected.insert(Coordinate {
        step: 2,
        x: 2,
        y: 0,
    });
    expected.insert(Coordinate {
        step: 3,
        x: 1,
        y: 1,
    });
    expected.insert(Coordinate {
        step: 4,
        x: 2,
        y: 2,
    });
    let input = "R2,U2".to_owned().split(",").map(String::from).collect();
    assert_eq!(to_coordinates(input), expected);
}

#[test]
fn test_part_one() {
    let input = "R75,D30,R83,U83,L12,D49,R71,U7,L72\n\
                 U62,R66,U55,R34,D71,R55,D58,R83"
        .to_owned();
    assert_eq!(closest(input), Some(159));
    let input = "R98,U47,R26,D63,R33,U87,L62,D20,R33,U53,R51\n\
                 U98,R91,D20,R16,D67,R40,U7,R15,U6,R7"
        .to_owned();
    assert_eq!(closest(input), Some(135));
}
